'use strict';

// ** Dependencies
const Promise = require('bluebird');
const Action = require('./actions');
const EventEmitter = require('./events').EventEmitter;
const functions = require('./functions');

class Process extends EventEmitter {
    constructor(name, options) {
        super();

        this._name = name;
        this._options = options;
        this._steps = [];
    }

    get name() {
        return this._name;
    }

    get run() {
        // Create a new action that run each process
        const self = this;
        const action = Action(self.name, function () {

            const steps = self._steps;

            // Track the position in the actions array to run
            let pos = 0;

            // Run the next action in the chain
            function next() {
                const args = functions.Arguments(arguments);

                // If we are done, then return the input
                if (pos === steps.length)
                    return args.shift();

                const step = steps[pos++];

                const type = step.type;
                const action = step.action;
                const options = step.options || {};

                const run = () => {
                    const getParams = () => options.params
                        ? options.params(args, action)
                        : args;

                    const getResult = result => options.result
                        ? options.result(result, args, action)
                        : result;

                    return Promise.try(getParams)
                        .then(params => action.run.apply(action, params))
                        .then(result => getResult(result));
                };

                // Run the action, then move next
                return run()
                    .then(next);
            }

            // Run the first step
            return next.apply(self, arguments);
        });

        // Forward Process Events as Action Events
        action.onAny((event, data) => this.emit(event, data));

        return action.run;
    }

    addStep(type, action, options) {

        // ** Emit events as process events
        action.onAny((event, data) => this.emit(`action:${action.name}:${event}`, data));

        this._steps.push({
            type: type,
            action: action,
            options: options
        });

        return action;
    }

    then(name, handler, options) {
        const action = Action(name, handler, options);

        this.addStep('then', action, {
            params: (args, action) => args,
            result: (result, args, action) => result // Use input as results
        });

        return this;
    }

    tap(name, handler, options) {
        const action = Action(name, handler, options);

        this.addStep('tap', action , {
            params: (args, action) => args,
            result: (result, args, action) => args // Use input as results
        });

        return this;
    }
}

function createProcess(name, options) {

    const process = new Process();
    return process;
}

// ** Exports
module.exports = createProcess;